import {Value} from './index';

/**
 * Sass's [mixin type](https://sass-lang.com/documentation/values/mixins).
 *
 * @category Custom Function
 */
export class SassMixin extends Value {
  /**
   * It is not possible to construct a Sass mixin outside of Sass. Attempting to
   * construct one will throw an exception.
   */
  constructor();
}
